{% extends "layout.html" %}
{% block body %}
<h3>
    Welcome to GNS3.
</h3>
<ul>
    <li><a href="https://gns3.com">Website</a></li>
    <li><a href="http://api.gns3.net">API documentation</a></li>
    <li><a href="/upload">Upload images & backup</a></li>
</ul>
{% endblock %}
